@@ -3,6 +3,7 @@ Variogram class
 """
 import copy
 import warnings
+from typing import Iterable, Callable, Union, Tuple
 
 import numpy as np
 from pandas import DataFrame
@@ -101,7 +102,7 @@ class Variogram(object):
             scipy.spatial.distance.pdist. Additional parameters are not (yet)
             passed through to pdist. These are accepted by pdist for some of
             the metrics. In these cases the default values are used.
-        bin_func : str
+        bin_func : str | Callable | Iterable
             .. versionchanged:: 0.3.8
                 added 'fd', 'sturges', 'scott', 'sqrt', 'doane'
             .. versionchanged:: 0.3.9
@@ -131,13 +132,14 @@ class Variogram(object):
         normalize : bool
             Defaults to False. If True, the independent and dependent
             variable will be normalized to the range [0,1].
-        fit_method : str
+        fit_method : str | None
             .. versionchanged:: 0.3.10
                 Added 'ml' and 'custom'
 
             String identifying the method to be used for fitting the
-            theoretical variogram function to the experimental. More info is
-            given in the Variogram.fit docs. Can be one of:
+            theoretical variogram function to the experimental. If None
+            is passed, the fit does not run. More info is given in the
+            Variogram.fit docs. Can be one of:
 
                 * 'lm': Levenberg-Marquardt algorithm for unconstrained
                   problems. This is the faster algorithm, yet is the fitting of
@@ -227,6 +229,14 @@ class Variogram(object):
             If :func:`bin_func <skgstat.Variogram.set_bin_func>` is `'ward'`
             this keyword argument can switch from default mean aggregation to
             median aggregation for calculating the cluster centroids.
+        obs_sigma : int, float
+            .. versionadded:: 0.6.0
+
+            If set, the Variogram will use this sigma as the standard deviation
+            of the observations passed as values. Using a MonteCarlo simulation
+            the uncertainties are propagated into the experimental variogram.
+            If present, the plot will indicate the confidence interval as
+            error bars around the experimental variogram.
 
         """
         # Before we do anything else, make kwargs available
@@ -303,10 +313,13 @@ class Variogram(object):
         self._bin_func = None
         self._groups = None
         self._bins = None
+        self._bin_count = None
         self.set_bin_func(bin_func=bin_func)
 
-        # Needed for harmonized models
-        self.preprocessing(force=True)
+        # Needed for harmonized models, no need to
+        # run if there is no fit
+        if fit_method is not None:
+            self.preprocessing(force=True)
 
         # model can be a function or a string
         self._model = None
@@ -320,7 +333,7 @@ class Variogram(object):
         self.use_nugget = use_nugget
 
         # set the fitting method and sigma array
-        self.fit_method = fit_method
+        self._fit_method = fit_method
         self._fit_sigma = None
         self.fit_sigma = fit_sigma
 
@@ -335,6 +348,12 @@ class Variogram(object):
         # Note that fit() calls preprocessing
         self.fit(force=True)
 
+        # finally check if any of the uncertainty propagation kwargs are set
+        self._experimental_conf_interval = None
+        self._model_conf_interval = None
+        if 'obs_sigma' in self._kwargs:
+            self._propagate_obs_sigma()
+
     @property
     def coordinates(self):
         """Coordinates property
@@ -350,6 +369,31 @@ class Variogram(object):
         """
         return self._X.coords
 
+    @property
+    def metric_space(self):
+        r"""
+        .. versionadded:: 0.5.6
+        
+        :class:`MetricSpace <skgstat.MetricSpace>` representation of the
+        input coordinates. A :class:`MetricSpace <skgstat.MetricSpace>`
+        can be used to pass pre-calculated coordinates to other
+        :class:`Variogram <skgstat.Variogram>` instances.
+
+        Returns
+        -------
+        metric_space : skgstat.MetricSpace
+
+        See Also
+        --------
+        Variogram.coordinates : coordinate representation
+
+        """
+        return self._X
+
+    @metric_space.setter
+    def metric_space(self, value):
+        raise AttributeError('MetricSpace is read-only.')
+
     @property
     def dim(self):
         """
@@ -484,12 +528,15 @@ class Variogram(object):
     def bin_func(self, bin_func):
         self.set_bin_func(bin_func=bin_func)
 
-    def set_bin_func(self, bin_func: str):
+    def set_bin_func(self, bin_func: Union[str, Iterable, Callable[[np.ndarray, float, float], Tuple[np.ndarray, float]]]):
         r"""Set binning function
 
         Sets a new binning function to be used. The new binning method is set
-        by a string identifying the new function to be used. Can be one of:
-        ['even', 'uniform', 'fd', 'sturges', 'scott', 'sqrt', 'doane'].
+        by either a string identifying the new function to be used, or an
+        iterable containing the bin edges, or any function that can compute
+        bins from the distances, number of lags and maximum lag.
+        The string can be one of: ['even', 'uniform', 'fd',
+         'sturges', 'scott', 'sqrt', 'doane'].
         If the number of lag classes should be estimated automatically, it is
         recommended to use ' sturges' for small, normal distributed locations
         and 'fd' or 'scott' for large datasets, where 'fd' is more robust to
@@ -510,9 +557,13 @@ class Variogram(object):
             pass kwargs to the binning functions is now implemented as
             a instance method, to make it pickleable.
 
+        .. versionchanged:: 0.6.5
+            added iterable and function as arguments to allow for custom
+            bins.
+
         Parameters
         ----------
-        bin_func : str
+        bin_func : str | Iterable | Callable
             Can be one of:
 
                 * 'even'
@@ -616,35 +667,50 @@ class Variogram(object):
         """
         # handle strings
         if isinstance(bin_func, str):
+            fname = bin_func.lower()
             # switch the input
-            if bin_func.lower() == 'even':
+            if fname == 'even':
                 self._bin_func = binning.even_width_lags
 
-            elif bin_func.lower() == 'uniform':
+            elif fname == 'uniform':
                 self._bin_func = binning.uniform_count_lags
 
             # remove the n_lags if they will be adjusted on call
             else:
                 # reset lags for adjusting algorithms
-                if bin_func.lower() not in ('kmeans', 'ward', 'stable_entropy'):
+                if fname not in ('kmeans', 'ward', 'stable_entropy'):
                     self._n_lags = None
 
                 # use the wrapper for all but even and uniform
                 self._bin_func = self._bin_func_wrapper
+            bin_func_name = bin_func
 
-        elif callable(bin_func):  # pragma: no cover
+        elif isinstance(bin_func, Callable):  # pragma: no cover
             self._bin_func = bin_func
-            bin_func = 'custom'
+            bin_func_name = 'custom_func'
+
+        elif isinstance(bin_func, Iterable):
+            self._bin_func = None
+            bin_func_name = 'custom_bin_edges'
 
         else:
-            raise AttributeError('bin_func has to be of type string.')
+            raise AttributeError('bin_func has to be of type string, iterable or callable.')
 
         # store the name
-        self._bin_func_name = bin_func
+        self._bin_func_name = bin_func_name
 
-        # reset groups and bins
+        # reset groups, bins and bin count
         self._groups = None
-        self._bins = None
+        self._bin_count = None
+
+        if isinstance(bin_func, str) or isinstance(bin_func, Callable):
+            self._bins = None
+        # if the input is an iterable with bin edges, we need to write self._bins here
+        else:
+            self._bins = bin_func
+            self._maxlag = max(bin_func)
+            self._n_lags = sum(1 for e in bin_func)
+
         self.cof, self.cov = None, None
 
     def _bin_func_wrapper(self, distances, n, maxlag):
@@ -661,10 +727,19 @@ class Variogram(object):
             return binning.ward(distances, n, maxlag)
 
         elif self._bin_func_name.lower() == 'stable_entropy':
-            return binning.stable_entropy_lags(distances, n, maxlag, **self._kwargs)
+            return binning.stable_entropy_lags(
+                distances,
+                n,
+                maxlag,
+                **self._kwargs
+            )
 
         else:
-            return binning.auto_derived_lags(distances, self._bin_func_name, maxlag)
+            return binning.auto_derived_lags(
+                distances,
+                self._bin_func_name,
+                maxlag
+            )
 
     @property
     def normalized(self):
@@ -710,6 +785,7 @@ class Variogram(object):
         self._bins = np.asarray(bins)
 
         # clean the groups as they are not valid anymore
+        self._bin_count = None
         self._groups = None
         self.cov = None
         self.cof = None
@@ -752,13 +828,21 @@ class Variogram(object):
         # if there are no errors, store the passed value
         self._n_lags_passed_value = n
 
-        # reset the groups
+        # reset the groups and bin count
         self._groups = None
+        self._bin_count = None
 
         # reset the fitting
         self.cof = None
         self.cov = None
 
+    @property
+    def bin_count(self):
+
+        if self._bin_count is None:
+            self._bin_count = np.fromiter((g.size for g in self.lag_classes()), dtype=int)
+        return self._bin_count
+
     @property
     def estimator(self):
         return self._estimator
@@ -939,7 +1023,7 @@ class Variogram(object):
         # handle sparse matrix
         if isinstance(self.distance_matrix, sparse.spmatrix):
             return self.triangular_distance_matrix.data
-        
+
         # Turn it back to triangular form not to have duplicates
         return squareform(self.distance_matrix)
 
@@ -956,7 +1040,10 @@ class Variogram(object):
         c = m.tocsc()
         c.data = c.indices
         rows = c.tocsr()
-        filt = sparse.csr.csr_matrix((m.indices < rows.data, m.indices, m.indptr), m.shape)
+        filt = sparse.csr.csr_matrix(
+            (m.indices < rows.data, m.indices, m.indptr),
+            m.shape
+        )
         return m.multiply(filt)
 
     @property
@@ -965,6 +1052,26 @@ class Variogram(object):
 
     @property
     def maxlag(self):
+        """
+        Maximum lag distance to be considered in this Variogram instance.
+        You can limit the distance at which point pairs are calcualted.
+        There are three possible ways how to do that, in absoulte lag units,
+        which is a number larger one. Secondly, a number ``0 < maxlag < 1``
+        can be set, which will use this share of the maximum distance as
+        maxlag. Lastly, a string can be set: ``'mean'`` and ``'median'``
+        for the mean or median value of the distance matrix.
+
+        Notes
+        -----
+        This setting is largely flexible, but all options except the
+        absolute limit in lag units need the full distance matrix to be
+        calculated. Hence, it does **not** speed up the calculation
+        of large distance matrices, just the estimation of the variogram.
+        Thus, if you pre-calcualte the distance matrix using
+        :class:`MetricSpace <skgstat.MetricSpace>`, only absoulte
+        limits can be used.
+
+        """
         return self._maxlag
 
     @maxlag.setter
@@ -972,10 +1079,11 @@ class Variogram(object):
         # reset fitting
         self.cof, self.cov = None, None
 
-        # remove bins
+        # remove bins, groups, and bin count
         self._bins = None
         self._groups = None
-        
+        self._bin_count = None
+
         # set new maxlag
         if value is None:
             self._maxlag = None
@@ -989,6 +1097,63 @@ class Variogram(object):
         else:
             self._maxlag = value
 
+    @property
+    def fit_method(self):
+        """
+        .. versionadded:: 0.6.2
+
+        Set the fit method to be used for this Variogram instance.
+        Possible values are:
+
+        * ``'trf'``   - Trust-Region Reflective (default)
+        * ``'lm'``    - Levenberg-Marquardt
+        * ``'ml'``    - Maximum Likelihood estimation
+        * `'manual'`` - Manual fitting by setting the parameters
+
+        .. versionchanged:: 0.6.6
+            Passing `None` will prevent the fitting procedure from
+            running.
+
+        See Also
+        --------
+        scipy.optimize.minimize
+        scipy.optimize.curve_fit
+        Variogram.fit
+
+        Notes
+        -----
+        The default method (TRF) is a bounded least squares method,
+        that sets constraints to the value space of all parameters.
+        All methods use an initial guess for all used parameters.
+        This is :func:`max(bins) <skgstat.Variogram.bins>` for
+        the range,
+        :func:`max(experimental) <skgstat.Variogram.experimental>`
+        for the sill, ``20`` for the Matérn smoothness, ``2`` for
+        the stable model shape and ``1`` for the nugget if used.
+
+        """
+        return self._fit_method
+
+    @fit_method.setter
+    def fit_method(self, value):
+        if value not in ('lm', 'ml', 'trf', 'manual'):
+            raise AttributeError(
+                "fit_method has to be one of ['lm', 'ml', 'trf', 'manual']"
+            )
+
+        # value is fine -check for manual does not drop the coefficients
+        elif value == 'manual':
+            self._fit_method = 'manual'
+
+        # otherwise - refit
+        else:
+            # set method
+            self._fit_method = value
+
+            # reset fit
+            self.cof = None
+            self.cov = None
+
     @property
     def fit_sigma(self):
         r"""Fitting Uncertainty
@@ -998,7 +1163,8 @@ class Variogram(object):
         cost function, which divides the residuals by their uncertainty.
 
         When setting fit_sigma, the array of uncertainties itself can be
-        given, or one of the strings: ['linear', 'exp', 'sqrt', 'sq', 'entropy'].
+        given, or one of the strings:
+        ['linear', 'exp', 'sqrt', 'sq', 'entropy'].
         The parameters described below refer to the setter of this property.
 
         .. versionchanged:: 0.3.11
@@ -1054,7 +1220,7 @@ class Variogram(object):
             if len(self._fit_sigma) == len(self._bins):
                 return self._fit_sigma
             else:
-                raise AttributeError('fit_sigma and bins need the same length.')
+                raise AttributeError('len(fit_sigma) != len(bins)')
 
         # linear function of distance
         elif self._fit_sigma == 'linear':
@@ -1228,15 +1394,15 @@ class Variogram(object):
                 scipy.optimize.leastsq function.
               * trf: Trust Region Reflective algorithm implemented in
                 scipy.optimize.least_squares(method='trf')
-              * 'ml': Maximum-Likelihood estimation. With the current implementation
-                only the Nelder-Mead solver for unconstrained problems is
-                implemented. This will estimate the variogram parameters from
-                a Gaussian parameter space by minimizing the negative
-                log-likelihood.
+              * 'ml': Maximum-Likelihood estimation. With the current
+                implementation only the Nelder-Mead solver for unconstrained
+                problems is implemented. This will estimate the variogram
+                parameters from a Gaussian parameter space by minimizing
+                the negative log-likelihood.
               * 'manual': Manual fitting. You can set the range, sill and
-                nugget either directly to the :func:`fit <skgstat.Variogram.fit>`
-                function, or as `fit_` prefixed keyword arguments on
-                Variogram instantiation.
+                nugget either directly to the
+                :func:`fit <skgstat.Variogram.fit>` function, or as
+                `fit_` prefixed keyword arguments on Variogram instantiation.
 
 
         sigma : string, array
@@ -1262,8 +1428,9 @@ class Variogram(object):
             old_params = self.describe()
 
         # delete the last cov and cof
-        self.cof = None
-        self.cov = None
+        if force:
+            self.cof = None
+            self.cov = None
 
         # if force, force a clean preprocessing
         self.preprocessing(force=force)
@@ -1390,9 +1557,6 @@ class Variogram(object):
             else:
                 self.cof = [r, s, n]
 
-        else:
-            raise ValueError("fit method has to be one of ['trf', 'lm', 'ml', 'custom']")
-
     def transform(self, x):
         """Transform
 
@@ -1456,14 +1620,13 @@ class Variogram(object):
             if kw["nugget"] != 0:
                 cof.append(kw["nugget"])
 
-        harmonize = False
         if not callable(model):
             if model == "harmonize":
                 raise ValueError("Please supply the actual harmonized model directly")
             else:
                 model = model.lower()
                 model = getattr(models, model)
-            
+
         if model.__name__ == "harmonize":
             code = """fitted_model = lambda x: model(x)"""
         else:
@@ -1511,7 +1674,7 @@ class Variogram(object):
             self._diff = np.abs(Vrow.data - Vcol.data)
         else:
             # Append a column of zeros to make pdist happy
-            # euclidean: sqrt((a-b)**2 + (0-0)**2) == sqrt((a-b)**2) == abs(a-b)
+            # euclidean: sqrt((a-b)**2 + (0-0)**2) == sqrt((a-b)**2)
             self._diff = pdist(
                 np.column_stack((v, np.zeros(len(v)))),
                 metric="euclidean"
@@ -1735,7 +1898,6 @@ class Variogram(object):
         self.preprocessing(force=force)
 
         # calculate the experimental variogram
-        _exp = self.experimental
         _bin = self.bins
 
         x = np.linspace(0, np.float64(np.nanmax(_bin)), n)
@@ -1746,6 +1908,22 @@ class Variogram(object):
 
         return x, self._model(x, *self.cof)
 
+    def _propagate_obs_sigma(self):
+        """
+        Observation uncertainties are propagated into the experimental
+        variogram.
+        """
+        # due to circular import, import here
+        from skgstat.util.uncertainty import propagate
+
+        # TODO: load and propagate percentile settings
+        self._experimental_conf_interval = propagate(
+            self,
+            'values',
+            self._kwargs['obs_sigma'],
+            use_bounds=True
+        )
+
     @property
     def residuals(self):
         """Model residuals
@@ -1878,7 +2056,6 @@ class Variogram(object):
 
         return mae
 
-
     @property
     def nrmse(self):
         r"""NRMSE
@@ -1977,8 +2154,16 @@ class Variogram(object):
         mx = np.nanmean(experimental)
 
         # calculate the single nash-sutcliffe terms
-        term1 = np.nansum(np.fromiter(map(lambda x, y: (x - y)**2, experimental, model), float))
-        term2 = np.nansum(np.fromiter(map(lambda x: (x - mx)**2, experimental), float))
+        term1 = np.nansum(np.fromiter(
+            map(lambda x, y: (x - y)**2, experimental, model),
+            float
+            )
+        )
+        term2 = np.nansum(np.fromiter(
+            map(lambda x: (x - mx)**2, experimental),
+            float
+            )
+        )
 
         return 1 - (term1 / term2)
 
@@ -2033,7 +2218,7 @@ class Variogram(object):
             The number of points to be included into the cross-validation.
             If None (default), all points will be used.
         metric : str
-            Metric used for cross-validation. 
+            Metric used for cross-validation.
             Can be root mean square error (rmse), mean squared error (mse)
             or mean absolute error (mae).
         seed : int
@@ -2053,7 +2238,12 @@ class Variogram(object):
         from skgstat.util import cross_validation
 
         if method == 'jacknife':
-            return cross_validation.jacknife(self, n=n, metric=metric, seed=seed)
+            return cross_validation.jacknife(
+                self,
+                n=n,
+                metric=metric,
+                seed=seed
+            )
         else:
             raise AttributeError(f"A method '{method}' is not implemented.")
 
@@ -2137,7 +2327,7 @@ class Variogram(object):
                 dist_func=str(self.dist_function),
                 bin_func=self._bin_func_name,
                 normalize=self.normalized,
-                fit_method=self.fit_method,
+                fit_method=self._fit_method,
                 fit_sigma=self.fit_sigma,
                 use_nugget=self.use_nugget,
                 maxlag=self.maxlag,
@@ -2159,9 +2349,16 @@ class Variogram(object):
     @property
     def parameters(self):
         """
-        Extract just the variogram parameters range, sill and nugget from the self.describe return
+        Extract just the variogram parameters range, sill
+        and nugget from the
+        :func:`describe <skgstat.Variogram.describe>` output.
+
+        Returns
+        -------
+        params : list
+            [range, sill, nugget] for most models and
+            [range, sill, shape, nugget] for matern and stable model.
 
-        :return:
         """
         d = self.describe()
         if 'error' in d:
@@ -2260,6 +2457,9 @@ class Variogram(object):
         >> cond_pos Variogram.coordinates.T
 
         """
+        # first fit again
+        self.fit()
+
         return skgstat_to_gstools(self, **kwargs)
 
     def to_gs_krige(self, **kwargs):
@@ -2299,6 +2499,8 @@ class Variogram(object):
         gstools.Krige
 
         """
+        # fit 
+        self.fit()
         return skgstat_to_krige(self, **kwargs)
 
     def plot(self, axes=None, grid=True, show=True, hist=True):
@@ -2342,9 +2544,21 @@ class Variogram(object):
         used_backend = plotting.backend()
 
         if used_backend == 'matplotlib':
-            return plotting.matplotlib_variogram_plot(self, axes=axes, grid=grid, show=show, hist=hist)
+            return plotting.matplotlib_variogram_plot(
+                self,
+                axes=axes,
+                grid=grid,
+                show=show,
+                hist=hist
+            )
         elif used_backend == 'plotly':
-            return plotting.plotly_variogram_plot(self, fig=axes, grid=grid, show=show, hist=hist)
+            return plotting.plotly_variogram_plot(
+                self,
+                fig=axes,
+                grid=grid,
+                show=show,
+                hist=hist
+            )
 
         # if we reach this line, somethings wrong with plotting backend
         raise ValueError('The plotting backend has an undefined state.')
@@ -2380,9 +2594,19 @@ class Variogram(object):
         used_backend = plotting.backend()
 
         if used_backend == 'matplotlib':
-            return plotting.matplotlib_variogram_scattergram(self, ax=ax, show=show, **kwargs)
+            return plotting.matplotlib_variogram_scattergram(
+                self,
+                ax=ax,
+                show=show,
+                **kwargs
+            )
         elif used_backend == 'plotly':
-            return plotting.plotly_variogram_scattergram(self, fig=ax, show=show, **kwargs)
+            return plotting.plotly_variogram_scattergram(
+                self,
+                fig=ax,
+                show=show,
+                **kwargs
+            )
 
         # if we reach this line, somethings wrong with plotting backend
         raise ValueError('The plotting backend has an undefined state.')
@@ -2437,9 +2661,19 @@ class Variogram(object):
         used_backend = plotting.backend()
 
         if used_backend == 'matplotlib':
-            return plotting.matplotlib_location_trend(self, axes=axes, show=show, **kwargs)
+            return plotting.matplotlib_location_trend(
+                self,
+                axes=axes,
+                show=show,
+                **kwargs
+            )
         elif used_backend == 'plotly':
-            return plotting.plotly_location_trend(self, fig=axes, show=show, **kwargs)
+            return plotting.plotly_location_trend(
+                self,
+                fig=axes,
+                show=show,
+                **kwargs
+            )
 
         # if we reach this line, somethings wrong with plotting backend
         raise ValueError('The plotting backend has an undefined state.')
@@ -2474,9 +2708,19 @@ class Variogram(object):
         used_backend = plotting.backend()
 
         if used_backend == 'matplotlib':
-            return plotting.matplotlib_dd_plot(self, ax=ax, plot_bins=plot_bins, show=show)
+            return plotting.matplotlib_dd_plot(
+                self,
+                ax=ax,
+                plot_bins=plot_bins,
+                show=show
+            )
         elif used_backend == 'plotly':
-            return plotting.plotly_dd_plot(self, fig=ax, plot_bins=plot_bins, show=show)
+            return plotting.plotly_dd_plot(
+                self,
+                fig=ax,
+                plot_bins=plot_bins,
+                show=show
+            )
 
         # if we reach this line, somethings wrong with plotting backend
         raise ValueError('The plotting backend has an undefined state.')
@@ -2490,7 +2734,7 @@ class Variogram(object):
         try:
             _name = self._model.__name__
             _b = int(len(self.bins))
-        except:
+        except Exception:
             return "< abstract Variogram >"
         return "< %s Semivariogram fitted to %d bins >" % (_name, _b)
 
@@ -2514,8 +2758,8 @@ class Variogram(object):
         _nugget = np.NaN if 'error' in par else par['nugget']
 
         s = "{0} Variogram\n".format(par['model'])
-        s+= "-" * (len(s) - 1) + "\n"
-        s+="""Estimator:         %s
+        s += "-" * (len(s) - 1) + "\n"
+        s += """Estimator:         %s
         \rEffective Range:   %.2f
         \rSill:              %.2f
         \rNugget:            %.2f
